import ebcdic
from dearpygui import dearpygui as dpg
from zosedit.models import Dataset, Job, Spool
from zosedit.constants import tempdir
from zosedit.zftp import zFTP
from zosedit.gui.dialog import dialog
from pathlib import Path
from datetime import datetime


class Tab:

    def __init__(self, *, ftp: zFTP = None, dataset: Dataset = None, job: Job = None):
        self.ftp = ftp
        self.dataset = dataset
        self.job = job
        self.dirty = False
        self.uuid = None
        self.label = None
        self._line_height = dpg.get_text_size('')[1] + 18

        if dataset:
            self.build_dataset_tab()
        elif job:
            self.build_job_tab()
        else:
            self.uuid = dpg.add_tab(label='   ', closable=False, parent='editor_tab_bar')

    def build_dataset_tab(self):
        dataset: Dataset = self.dataset

        # Clear existing tab / create new tab
        if self.uuid:
            for child in dpg.get_item_children(self.uuid)[1]:
                dpg.delete_item(child)
        else:
            self.uuid = dpg.add_tab(label=dataset.name, closable=True, parent='editor_tab_bar')
            self.label = dataset.name

        # Setup tab
        dpg.bind_item_theme(self.uuid, 'dataset_tab_theme')
        dpg.set_value('editor_tab_bar', self.uuid)
        with dpg.child_window(parent=self.uuid, height=self._line_height, border=False, horizontal_scrollbar=True):
            dpg.add_text(str(dataset))

        # Get file content
        if dataset.new:
            self.mark_dirty()
        else:
            status = dpg.add_text('Downloading...', parent=self.uuid)
            self.ftp.download(dataset)
            dpg.delete_item(status)
        content = dataset.local_path.read_text(errors='replace')
        lines = [line.rstrip() for line in content.split('\n')]
        text = '\n'.join(lines)

        # Create editor
        self.editor = dpg.add_input_text(
            parent=self.uuid,
            default_value=text,
            multiline=True,
            width=-1,
            height=-1,
            callback=self.mark_dirty,
            tab_input=True,
            user_data=self)


    def build_job_tab(self):
        label = f'{self.job.id} ({self.job.name})'

        # Clear existing tab / create new tab
        if self.uuid:
            for child in dpg.get_item_children(self.uuid)[1]:
                dpg.delete_item(child)
        else:
            self.uuid = dpg.add_tab(label=label, closable=True, parent='editor_tab_bar')
            self.label = label

        # Setup tab
        dpg.bind_item_theme(self.uuid, 'job_tab_theme')
        dpg.set_value('editor_tab_bar', self.uuid)

        # Info/status
        with dpg.child_window(parent=self.uuid, height=self._line_height, border=False, horizontal_scrollbar=True):
            dpg.add_text(str(self.job))
        status = dpg.add_text('Downloading spool...', parent=self.uuid)

        # Create spool dropdowns
        self.spool_headers = []
        for spool in self.ftp.list_spools(self.job):
            header = dpg.add_collapsing_header(before=status, label=spool.ddname, parent=self.uuid)
            self.spool_headers.append(header)
            with dpg.item_handler_registry() as reg:
                dpg.add_item_toggled_open_handler(callback=self._populate_spool, user_data=(header,spool))
            dpg.bind_item_handler_registry(header, reg)
        dpg.delete_item(status)


    def _populate_spool(self, sender, data, user_data):
        header, spool = user_data
        # for h in self.spool_headers:
        #     dpg.set_value(h, h == header)

        if dpg.get_item_children(header)[1]: # Already populated
            return

        # Info/status
        with dpg.child_window(parent=header, height=self._line_height, border=False, horizontal_scrollbar=True):
            dpg.add_text(str(spool), indent=10)
        status = dpg.add_text('Downloading spool output...', parent=header, indent=10)
        if not self.ftp.download_spool(spool):
            dpg.set_value(status, 'Download failed')
            dpg.configure_item(status, color=(255, 255, 0))
            return
        dpg.delete_item(status)

        # Display spool
        text = spool.local_path.read_text()
        tw, th = dpg.get_text_size(text)

        with dpg.child_window(parent=header,
                              horizontal_scrollbar=True, border=False) as window:
            input_field = dpg.add_input_text(multiline=True,
                                             width=tw + 20,
                                             height=th + 20,
                                             default_value=text,
                                             readonly=True,)
            dpg.bind_item_theme(input_field, 'spool_input_theme')
        self.resize_spool_window(None, None, (header, window, input_field))

        # Resize window handler
        with dpg.item_handler_registry() as reg:
            dpg.add_item_toggled_open_handler(callback=self.resize_spool_window, user_data=(header, window, input_field))
        dpg.bind_item_handler_registry(header, reg)

    def resize_spool_window(self, sender, data, user_data):
        print(f'resize_spool_window {datetime.now()}')
        header, window, input_field = user_data
        width = dpg.get_item_rect_size(header)[0]
        text = dpg.get_value(input_field)
        tw, th = dpg.get_text_size(text)
        w = min(tw + 34, width - 15)
        h = min(th + 34, dpg.get_viewport_height() - 220)
        dpg.configure_item(window, width=w, height=h)

    def mark_dirty(self):
        dpg.configure_item(self.uuid, label=self.dataset.name + '*')
        self.dirty = True

    def mark_clean(self):
        dpg.configure_item(self.uuid, label=self.dataset.name)
        self.dirty = False
        self.dataset.new = False

    def __repr__(self):
        return f"Tab({self.label} - {self.uuid})"


class Editor:

    def __init__(self, root):
        self.root = root
        self.tabs = []

    def build(self):
        dpg.add_tab_bar(tag='editor_tab_bar', reorderable=True, callback=self.on_tab_changed)

        with dpg.theme(tag='job_tab_theme'):
            with dpg.theme_component(dpg.mvTab):
                dpg.add_theme_color(dpg.mvThemeCol_Tab, (40, 70, 50, 255), category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_TabActive, (40, 140, 78, 255), category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_TabHovered, (30, 130, 68, 255), category=dpg.mvThemeCat_Core)

        with dpg.theme(tag='dataset_tab_theme'):
            with dpg.theme_component(dpg.mvTab):
                dpg.add_theme_color(dpg.mvThemeCol_Tab, (50, 60, 80, 255), category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_TabActive, (50, 60, 150, 255), category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_TabHovered, (30, 70, 130, 255), category=dpg.mvThemeCat_Core)

        with dpg.theme(tag='spool_input_theme'):
            with dpg.theme_component(dpg.mvInputText):
                dpg.add_theme_style(dpg.mvStyleVar_CellPadding, 0, 0, category=dpg.mvThemeCat_Core)
                dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 0, 0, category=dpg.mvThemeCat_Core)
                dpg.add_theme_style(dpg.mvStyleVar_WindowPadding, 0, 0, category=dpg.mvThemeCat_Core)



        with dpg.handler_registry():
            dpg.add_key_press_handler(dpg.mvKey_N, callback=self.new_file_keybind)
            dpg.add_key_press_handler(dpg.mvKey_S, callback=self.save_keybind)
            dpg.add_key_press_handler(dpg.mvKey_W, callback=self.close_tab_keybind)
            dpg.add_key_press_handler(dpg.mvKey_Tab, callback=self.switch_tab_keybind)

    def reset(self):
        tabs = [tab for tab in self.tabs]
        for tab in tabs:
            self.delete_tab(tab)
        self.tabs = []

    def on_tab_changed(self):
        self.update_internal_state()

    # Jobs
    def open_job(self, job: Job):
        tab = self.get_tab_by_job(job)
        if not tab:
            tab = Tab(ftp=self.root.zftp, job=job)
            self.tabs.append(tab)
        elif tab.dirty:
            self.switch_to_tab(tab)
        else:
            tab.build_job_tab()
        self.switch_to_tab(tab)

    # Files
    def open_file(self, dataset: Dataset):
        tab = self.get_tab_by_dataset(dataset.name)
        if not tab:
            tab = Tab(ftp=self.root.zftp, dataset=dataset)
            self.tabs.append(tab)
        elif tab.dirty:
            self.switch_to_tab(tab)
        else:
            tab.build_dataset_tab()
        self.switch_to_tab(tab)
        if dataset.new:
            self.get_current_tab().mark_dirty()

    def new_file(self, name=''):
        # Callback for creating a new file
        def create_file():
            dataset_name = dpg.get_value('new_file_dataset_input')

            type_ = dpg.get_value('new_file_type')
            type_ = 'PO' if type_ == 'Partitioned' else 'PS'
            format_ = dpg.get_value('new_file_format')
            format_ = 'FB' if format_ == 'Fixed Width' else 'VB'

            dataset = Dataset(dataset_name)
            dataset.new = True
            dataset.local_path = Path(tempdir, dataset_name)
            dataset.local_path.write_text('')
            dataset.reclength = dpg.get_value('new_file_record_length')
            dataset.recformat = format_
            dataset.type = type_

            if type_ == 'PO':
                self.root.zftp.mkdir(dataset)
            else:
                self.open_file(dataset)
            dpg.delete_item('new_file_dialog')

        # Close existing dialog
        if dpg.does_item_exist('new_file_dialog'):
            dpg.delete_item('new_file_dialog')

        def on_switch_format():
            format_ = dpg.get_value('new_file_format')
            dpg.configure_item('new_file_record_length', show=format_ == 'Fixed Width')

        # Create new dialog
        w, h = 500, 200
        with dialog(tag='new_file_dialog', width=w, height=h, label='New'):
            dpg.add_input_text(hint='Dataset Name',
                               tag='new_file_dataset_input',
                               width=-1,
                               uppercase=True,
                               on_enter=True,
                               callback=create_file,
                               default_value=name)
            dpg.add_combo(label='Type',
                                 items=('Sequential', 'Partitioned'),
                                 tag='new_file_type',
                                 default_value='Sequential',
                                 width=120)
            dpg.add_combo(label='Record Format',
                                 items=('Fixed Width', 'Variable Width'),
                                 tag='new_file_format',
                                 default_value='Fixed Width',
                                 width=120,
                                 callback=on_switch_format)
            dpg.add_input_int(label='Record Length', tag='new_file_record_length',
                              default_value=80, min_value=1, max_value=32767, step=0,
                              width=120)

            dpg.add_spacer(height=10)
            with dpg.group(horizontal=True):
                dpg.add_button(label='Create', callback=create_file, width=w / 2 - 13)
                dpg.add_button(label='Cancel', callback=lambda: dpg.delete_item('new_file_dialog'), width=w / 2 - 12)

            dpg.focus_item('new_file_dataset_input')


    def save_open_file(self):
        tab = self.get_current_tab()
        if not tab or not tab.dataset:
            return

        if tab.dirty:
            text: str = dpg.get_value(tab.editor)

            result = []
            pad_to = tab.dataset.reclength if tab.dataset.recformat == 'FB' else tab.dataset.reclength - 4
            for line in text.split('\n'):
                result.append(line.ljust(pad_to))
            text = ''.join(result)

            tab.dataset.local_path.write_bytes(text.encode('cp1047'))

            if not self.root.zftp.upload(tab.dataset):
                return
            tab.mark_clean()

            current_search = dpg.get_value('explorer_dataset_input')
            if current_search and current_search in tab.dataset.name:
                self.root.explorer.refresh_datasets()

    # Tabs
    def switch_to_tab(self, tab: Tab):
        dpg.set_value('editor_tab_bar', tab.uuid)

    def cycle_tabs(self, direction: int):
        self.update_internal_state()
        tabs = [tab.uuid for tab in self.tabs]
        tab = dpg.get_value('editor_tab_bar')
        index = tabs.index(tab) + direction
        index = index % len(tabs)
        tab = tabs[index]
        dpg.set_value('editor_tab_bar', tab)

    def get_current_tab(self) -> Tab:
        tab = dpg.get_value('editor_tab_bar')
        return self.get_tab_by_id(tab)

    def get_tab_by_job(self, job: Job) -> Tab:
        matching_tabs = [tab for tab in self.tabs if tab.job and tab.job.id == job.id]
        if len(matching_tabs) == 0:
            return None
        return matching_tabs.pop()

    def get_tab_by_dataset(self, dataset: str) -> Tab:
        matching_tabs = [tab for tab in self.tabs if tab.dataset and tab.dataset.name == dataset]
        if len(matching_tabs) == 0:
            return None
        return matching_tabs.pop()

    def get_tab_by_id(self, id: int):
        matching_tabs = [tab for tab in self.tabs if tab.uuid == id]
        if len(matching_tabs) == 0:
            return None
        return matching_tabs.pop()

    def save_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.save_open_file()

    def switch_tab_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.cycle_tabs(-1 if dpg.is_key_down(dpg.mvKey_Shift) else 1)

    def new_file_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.new_file()

    def close_tab_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            tab = self.get_current_tab()
            if tab:
                self.delete_tab(tab)

    def delete_tab(self, tab: Tab):
        dpg.delete_item(tab.uuid)
        self.tabs.remove(tab)

    def close_tab_by_dataset(self, dataset: Dataset):
        tab = self.get_tab_by_dataset(dataset.name)
        if tab:
            self.delete_tab(tab)

    def update_internal_state(self):
        try:
            children = dpg.get_item_children('editor_tab_bar')[1]
            _tabs = [tab for tab in self.tabs if tab.uuid in children]
            for tab in _tabs:
                if not dpg.is_item_visible(tab.uuid):
                    self.delete_tab(tab)
            _tabs.sort(key=lambda x: dpg.get_item_rect_min(x.uuid)[0])
            self.tabs = _tabs
        except Exception as e:
            pass






import re
import ebcdic
from typing import Literal
from ftplib import FTP
from pathlib import Path
from tempfile import NamedTemporaryFile
from dearpygui import dearpygui as dpg
from zosedit.constants import tempdir
from traceback import format_exc
from textwrap import indent
from .models import Dataset, Job, Spool
from zosedit.gui.dialog import dialog
from . import constants
from time import time


def waits(func):
    def wrapper(self, *args, **kwargs):
        self.waiting = True
        self.wait_start = dpg.get_frame_count()
        try:
            result = func(self, *args, **kwargs)
        except Exception as e:
            self.waiting = False
            raise e
        self.waiting = False
        return result
    return wrapper

class zFTP:

    KEEP_ALIVE_INTERVAL = 60

    def __init__(self, root):
        self.root = root
        self.host = None
        self.user = None
        self.password = None
        self.waiting = False
        self.wait_start = 0
        self.ftp = None
        self.last_keep_alive = time()

    def keep_alive(self):
        if time() - self.last_keep_alive > self.KEEP_ALIVE_INTERVAL:
            self.check_alive()
            self.last_keep_alive = time()

    ### Datasets
    @waits
    def list_datasets(self, search_string: str):
        files = []
        try:
            self.set_ftp_vars('SEQ')
            self.ftp.dir(search_string, files.append)
        except Exception as e:
            if '550' in str(e):
                return []
            print('Error listing datasets')
            print(indent(format_exc(), '    '))
            self.show_error(f'Error listing datasets:\n{e}')
            return []
        datasets = [Dataset(file_) for file_ in set(files[1:])]
        datasets = sorted(datasets, key=lambda x: (x.is_partitioned(), x.name, x.volume or ''))
        return datasets

    @waits
    def get_members(self, dataset: Dataset):
        members = []
        def append(line):
            members.append(line.split()[0])
        try:
            self.set_ftp_vars('SEQ', VOLUME=dataset.volume)
            self.ftp.dir(f"'{dataset.name}(*)'", append)
        except Exception as e:
            print('Error getting members for', dataset.name)
            print(indent(format_exc(), '    '))
            print(e)
        dataset._populated = True
        return members[1:] if members else []

    @waits
    def download(self, dataset: Dataset):
        raw_data = []
        # Download file
        def write(data):
            raw_data.append(data)

        try:
            self.set_ftp_vars('SEQ', VOLUME=dataset.volume)
            self.ftp.retrlines(f"RETR '{dataset.name}'", write)
            content = '\n'.join(raw_data)
            path = tempdir / dataset.name
            path.write_text(content, errors='replace')
            dataset.local_path = path
            return True
        except Exception as e:
            self.show_error(f'Error downloading dataset {dataset.name}:\n{e}')
            return False

    @waits
    def mkdir(self, dataset: Dataset):
        try:
            self.set_ftp_vars('SEQ', RECFM=dataset.recformat, LRECL=dataset.reclength, BLKSIZE=dataset.block_size)
            self.ftp.mkd(f"'{dataset.name}'")
        except Exception as e:
            self.show_error(f'Error creating partitioned dataset:\n{e}')
            return

    @waits
    def upload(self, dataset: Dataset):
        try:
            if dataset.member:
                self.set_ftp_vars('SEQ')
            else:
                self.set_ftp_vars('SEQ', RECFM=dataset.recformat, LRECL=dataset.reclength, BLKSIZE=dataset.block_size)
            self.ftp.storbinary(f"STOR '{dataset.name}'", dataset.local_path.open('rb'))
        except Exception as e:
            self.show_error(f'Error uploading dataset:\n{e}')
            return False
        return True

    @waits
    def delete(self, dataset: Dataset):
        try:
            self.set_ftp_vars('SEQ', VOLUME=dataset.volume)
            self.ftp.delete(f"'{dataset.name}'")
            print('Deleted', dataset.name)
        except Exception as e:
            self.show_error(f'Error deleting dataset:\n{e}')
            return False
        return True

    ### Jobs
    @waits
    def submit_job(self, dataset: Dataset, download=True):
        try:
            if download and not self.download(dataset):
                return False
            path = dataset.local_path
            self.set_ftp_vars('JES')
            response = self.ftp.storlines(f"STOR '{dataset.name}'", path.open('rb'))
            self.show_response(response)
        except Exception as e:
            self.show_error(f'Error submitting job:\n{e}')
            return False
        return True

    @waits
    def operator_command_prompt(self):
        def _submit_command():
            try:
                jcl = constants.OPERCMD_JCL.format(
                    name=dpg.get_value('operator_command_job_name').ljust(10),
                    params=dpg.get_value('operator_command_job_params'),
                    command=dpg.get_value('operator_command_input')
                )

                with NamedTemporaryFile(delete=False) as f:
                    path = Path(f.name)
                    path.write_text(jcl)

                self.set_ftp_vars('JES')
                response = self.ftp.storlines(f"STOR 'ZEDITOPR'", path.open('rb'))

                path.unlink()
                dpg.delete_item('operator_command_prompt')
                self.show_response(response)
            except Exception as e:
                dpg.delete_item('operator_command_prompt')
                self.show_error(f'Error submitting operator command:\n{e}')
                return

        w, h = 420, 150
        with dialog(tag='operator_command_prompt', label='Operator Command', width=w, height=h, modal=False):
            dpg.add_input_text(label='Command', tag='operator_command_input', hint='S <JOBNAME>',
                               on_enter=True, callback=_submit_command)
            dpg.add_spacer(height=5)

            with dpg.collapsing_header(label="Advanced"):
                dpg.add_input_text(label='Job Name', tag='operator_command_job_name', default_value='ZEDITOPR')
                dpg.add_input_text(label='Job Card Params', tag='operator_command_job_params',
                                   default_value='CLASS=A,MSGCLASS=X,MSGLEVEL=(1,1),NOTIFY=&SYSUID')

            dpg.add_button(label='Submit', callback=_submit_command)

    @waits
    def list_jobs(self, name=None, id=None, owner=None):
        name = name or '*'
        owner = owner or '*'
        id = id or '*'
        raw_data: list[str] = []
        try:
            self.set_ftp_vars(f'JES', JESJOBNAME=name, JESOWNER=owner, JESENTRYLIMIT=1000)
            self.ftp.dir(id, raw_data.append)
        except Exception as e:
            if '550' in str(e):
                return []
            self.show_error(f'Error listing jobs:\n{e}')
            return []

        # If only a single job is returned it provides a different format
        if '--------' in raw_data:
            raw_data = ['', raw_data[1] + '  ' + raw_data[-1]]

        result = [Job(job_str) for job_str in raw_data[1:]]
        result.sort(key=lambda job: (job.rc == 'Active', job.id), reverse=True)
        return result

    @waits
    def download_spools(self, job: Job):
        spools = self.list_spools(job)

        self.set_ftp_vars('JES')
        exceptions = []
        for spool in spools:
            spool_name = f'{job.id}.{spool.id}'
            try:
                path = tempdir / f'{job.id}-{spool.ddname}.txt'
                lines = []
                self.ftp.retrlines(f"RETR {spool_name}", lines.append)
                path.write_text('\n'.join(lines))
                spool.local_path = path
                yield spool
            except Exception as e:
                exceptions.append((spool, e))
                continue

        errors = []
        for spool, exception in exceptions:
            errors.append(f'Error downloading spool "{spool}":\n    {exception}')
        if errors:
            self.show_error('\n'.join(errors))

    @waits
    def download_spool(self, spool: Spool) -> bool:
        try:
            path = tempdir / f'{spool.id}.txt'
            lines = []
            self.set_ftp_vars('JES')
            self.ftp.retrlines(f"RETR {spool.job.id}.{spool.id}", lines.append)
            path.write_text('\n'.join(lines))
            spool.local_path = path
            return True
        except Exception as e:
            self.show_error(f'Error downloading spool {spool.job.id}.{spool.ddname}:\n{e}')
            return False

    @waits
    def list_spools(self, job: Job):
        raw_data: list[str] = []
        try:
            self.set_ftp_vars('JES')
            self.ftp.dir(job.id, raw_data.append)
        except Exception as e:
            self.show_error(f'Error listing spool outputs:\n{e}')
            return []

        return [Spool(spool_str, job) for spool_str in raw_data[4:-1]]

    ### Dialogs
    def show_error(self, message):
        print(indent(message, '    '))
        print(format_exc())
        with dialog(label='FTP Error', tag='error', autosize=True):
            dpg.add_text(message, color=(255, 0, 0))

    def show_response(self, response):
        with dialog(label='FTP Response', tag='ftp_response', width=300, height=150):
            dpg.add_text(response)
            match = re.search(r'(J\d+|JOB\d+)', response)
            if match:
                id = match.group(0)
                dpg.add_button(label=f'Open Job {id}',
                               width=-1,
                               callback=self._open_job_by_id,
                               user_data=id)
        print(response)

    def _open_job_by_id(self, sender, data, id):
        dpg.delete_item('ftp_response')
        job = self.list_jobs(id=id)[0]
        self.root.editor.open_job(job)

    ### Connection
    @waits
    def connect(self, host=None, user=None, password=None):
        host = host or self.host
        user = user or self.user
        password = password or self.password
        print(f'Connecting: {user}@{host}')
        self.ftp = FTP(host or self.host)
        self.ftp.login(user=user or self.user, passwd=password or self.password)
        self.host = host
        self.user = user
        self.password = password
        self.ftp.set_debuglevel(2)

        return True

    @waits
    def check_alive(self):
        try:
            if self.ftp:
                self.ftp.voidcmd('NOOP')
        except Exception as e:
            self.quit()
            self.connect()

    @waits
    def quit(self):
        try:
            if self.ftp:
                self.ftp.quit()
        except Exception as e:
            print('Error quitting')
            print(indent(format_exc(), '    '))


    def set_ftp_vars(self, mode=Literal['SEQ', 'JES', 'SQL'], **kwargs):
        self.check_alive()
        args = ' '.join(f"{key}={value}" for key, value in kwargs.items())
        self.ftp.sendcmd(f'SITE RESET')
        self.ftp.sendcmd(f'SITE FILETYPE={mode} {args}')



from dearpygui import dearpygui as dpg
import zosedit.gui.explorer as explorer
import zosedit.gui.editor as editor
from zosedit.gui.dialog import dialog

from zosedit.constants import tempdir
from zosedit.zftp import zFTP

import platform

if platform.system() == 'Windows':
    from os import startfile

class Root:

    def __init__(self):
        self.zftp = None
        self.explorer = explorer.Explorer(self)
        self.editor = editor.Editor(self)
        self.zftp = zFTP(self)

    def start(self):
        dpg.create_context()
        dpg.create_viewport(title='z/OS Edit', resizable=True)

        with dpg.window(label="Main", tag='win_main') as main:
            with dpg.menu_bar():
                with dpg.menu(label="File", tag='file_menu'):
                    dpg.add_menu_item(label="New", shortcut="Ctrl+N", callback=self.editor.new_file)
                    dpg.add_menu_item(label="Save", shortcut="Ctrl+S", callback=self.editor.save_open_file)
                    if platform.system() == 'Windows':
                        dpg.add_menu_item(label="Open Data Directory", callback=self.open_data_directory)
                with dpg.menu(label="Run", tag='run_menu'):
                    dpg.add_menu_item(label="Command", shortcut="F1", callback=self.zftp.operator_command_prompt)
                    # dpg.add_menu_item(label="Submit", shortcut="F5", callback=self.editor.submit_open_file)
                with dpg.menu(label="Session", tag='session_menu'):
                    dpg.add_menu_item(label="Login", callback=self.login)
                    dpg.add_menu_item(label="Logout", callback=self.logout)
                # with dpg.menu(label='Settings'):
                #     dpg.add_menu_item(label='Show Style Editor', callback=dpg.show_style_editor)

            with dpg.handler_registry():
                dpg.add_key_press_handler(dpg.mvKey_F1, callback=self.zftp.operator_command_prompt)

            with dpg.group(horizontal=True):
                with dpg.child_window(label="Explorer", width=375, height=-1, tag='win_explorer', border=False):
                    self.explorer.build()

                with dpg.child_window(label="Editor", menubar=False, tag='win_editor', border=False,
                                    width=-1, height=-1, horizontal_scrollbar=True):
                    self.editor.build()

        self.login()

        dpg.set_primary_window(main, True)
        dpg.setup_dearpygui()
        dpg.show_viewport()

        while dpg.is_dearpygui_running():
            if self.zftp.waiting:
                self.waiting_animation()
            elif dpg.does_item_exist('overlay'):
                dpg.delete_item('overlay')
            dpg.render_dearpygui_frame()
        dpg.destroy_context()

    def waiting_animation(self):
        if not dpg.does_item_exist('overlay'):
            dpg.add_viewport_drawlist(tag='overlay')

        width = 100
        margin = 30
        xmin = dpg.get_viewport_width() - width - margin
        xmax = dpg.get_viewport_width() - margin
        y = 10
        dpg.draw_line((xmin, y), (xmax, y), color=(37, 37, 38), parent='overlay')
        x1 = ((dpg.get_frame_count() - self.zftp.wait_start)) % width
        x2 = ((dpg.get_frame_count() - self.zftp.wait_start) + 10) % width
        if (x2 < x1):
            x2 = xmax - xmin
        # dpg.draw_text((x, 15), '...', parent='overlay')
        dpg.draw_line((x1 + xmin, y), (x2 + xmin, y), parent='overlay')

    def logout(self):
        self.zftp.quit()
        self.explorer.reset()
        self.editor.reset()
        self.login()

    def login(self):
        def _login():
            host = dpg.get_value('settings_host_input')
            username = dpg.get_value('settings_username_input')
            password = dpg.get_value('settings_password_input')
            if self.zftp:
                dpg.set_value('login_status', 'Closing existing connection...')
                self.zftp.quit()

            dpg.set_value('login_status', f'Connecting to {host}...')
            try:
                self.zftp.connect(host, username, password)
            except Exception as e:
                dpg.set_value('login_status', f'Error connecting: {e}')
                return

            dpg.delete_item('login_dialog')
            dpg.set_value('explorer_dataset_input', username)
            self.explorer.refresh_datasets()

        w, h = 420, 150
        # Create new dialog
        with dialog(tag='login_dialog', label='Login', width=w, height=h):
            kwargs = {'on_enter': True, 'callback': _login, 'width': -1}
            dpg.add_input_text(hint='Host', tag='settings_host_input', default_value='QAZOS205', **kwargs)
            dpg.add_input_text(hint='Username', tag='settings_username_input', uppercase=True, **kwargs)
            dpg.add_input_text(hint='Password', tag='settings_password_input', password=True, uppercase=True, **kwargs)
            with dpg.group(horizontal=True):
                bw = w // 2 - 12
                dpg.add_button(label='Login', callback=_login, width=bw)
                dpg.add_button(label='Cancel', callback=lambda: dpg.delete_item('login_dialog'), width=bw -1)
            dpg.add_text('', tag='login_status')
            dpg.focus_item('settings_username_input')


    def open_data_directory(self):
        startfile(tempdir)

root = Root()

def main():
    root.start()

if __name__ == '__main__':
    main()



from dearpygui import dearpygui as dpg
from contextlib import contextmanager

@contextmanager
def dialog(*, tag=None, popup=False, modal=True, width=0, height=0, **kwargs):
    if tag and dpg.does_item_exist(tag):
        dpg.delete_item(tag)
    window = dpg.add_window(tag=tag, popup=popup, modal=modal, width=width, height=height, **kwargs)
    dpg.push_container_stack(window)
    yield
    dpg.pop_container_stack()
    w, h = dpg.get_item_rect_size(window)
    w = width or w
    h = height or h
    vw, vh = dpg.get_viewport_width(), dpg.get_viewport_height()
    dpg.set_item_pos(window, (vw/2-w/2, vh/2-h/2))





from pathlib import Path
from tempfile import gettempdir


tempdir = Path(gettempdir()) / 'zosedit'
tempdir.mkdir(exist_ok=True)

OPERCMD_JCL = '''
//{name} JOB {params}
//CMD       EXEC PGM=IKJEFT01
//  COMMAND  '{command}'
'''.strip()



import re
from pathlib import Path

class Dataset:
    cols = 'volume', 'unit', 'date', 'ext', 'used', 'recformat', 'reclength', 'block_size', 'type', 'name'

    def __init__(self, string, member: str = None):
        self.string = string
        self.member = member
        self.new = False
        self.local_path : Path = None

        self.volume: str = None
        self.unit: int = None
        self.date: str = None
        self.ext: int = None
        self.used: int = None
        self.recformat: str = None
        self.reclength: int = None
        self.block_size: int = None
        self.type: str = None
        self.name: str = None
        self._populated = False

        try:
            data = string.split()
            if len(data) != len(self.cols):
                self.name = data[-1].replace("'", "")
                return

            for col, value in zip(self.cols, data):
                setattr(self, col, value)

            self.name = self.name.replace("'", "")
            if self.member:
                self.name = f"{self.name}({self.member})"

            self.reclength = int(self.reclength)
        except Exception as e:
            print('Error parsing dataset:', e)
            print(string)

    def properties(self) -> dict:
        cols = sorted(self.cols)
        return {col: getattr(self, col) for col in cols}

    def is_partitioned(self):
        return self.type == 'PO'

    def __repr__(self):
        attrs = ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)
        return f"Dataset({attrs})"

    def __str__(self):
        return ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)

    def __call__(self, member: str) -> 'Dataset':
        if not member:
            return self
        dataset = Dataset(self.string, member)
        return dataset


class Job:
    cols = 'name', 'id', 'owner', 'status', 'class', 'rc', 'spool_count'

    def __init__(self, string):
        self.string = string

        self.name: str = None
        self.id: str = None
        self.owner: str = None
        self.status: str = None
        self.class_: str = None
        self.rc: int = None
        self.spool_count: int = None
        try:
            weirdness = re.search(r'\(([^)]*)\)', string)
            if weirdness:
                group = weirdness.group()
                string = string.replace(group, group.replace(' ', '_'))
            string = string.replace('RC unknown', '?')

            for col, value in zip(self.cols, string.split()):
                setattr(self, col, value)

            if self.rc:
                if '=' in self.rc:
                    self.rc = self.rc.split('=')[1]
                    if self.rc.isdecimal():
                        self.rc = int(self.rc)
                else:
                    self.rc = self.rc.replace('_', '').replace('(', '').replace(')', '')
            elif self.status == 'ACTIVE':
                self.rc = self.status.capitalize()
            if self.spool_count is not None:
                self.spool_count = int(self.spool_count.split()[0])
        except Exception as e:
            print('Error parsing job:', e)
            print(string)

    def theme(self):
        if self.status == 'ACTIVE':
            return 'active'
        rc = '?' if self.rc is None else self.rc
        if rc == 0:
            return 'success'
        return 'error'

    def read(self, *args, **kwargs):
        return None

    def __repr__(self):
        attrs = ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)
        return f"Job({attrs})"

    def __str__(self):
        return ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)

class Spool:

    cols = 'id', 'stepname', 'procstep', 'c', 'ddname', 'byte_count'

    def __init__(self, string: str, job: Job):
        self.string = string
        self.job = job

        self.id: str = None
        self.stepname: str = None
        self.procstep: str = None
        self.c: str = None
        self.ddname: str = None
        self.byte_count: int = None
        try:
            data = string.split()
            if len(data) == len(self.cols) - 1:
                data.insert(3, '')

            for col, value in zip(self.cols, data):
                setattr(self, col, value)

            if self.byte_count is not None:
                self.byte_count = float(self.byte_count)
        except Exception as e:
            print('Error parsing spool:', e)
            print(string)

    def __repr__(self):
        attrs = ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)
        return f"Spool({self.job.id}, {attrs})"

    def __str__(self):
        return ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)



'''
To build the package, run:
    pip install --upgrade build;
    rm dist/*; python -m build

The built package will be in the dist/ directory

To upload the package to PyPI, run:
    pip install --upgrade twine;
    twine upload dist/*
'''

from setuptools import setup, find_packages
from pathlib import Path

readme = Path(__file__).parent / 'README.md'

setup(
    name='zosedit',
    version='0.0.14',
    description='FTP-based MVS Dataset Editor',
    long_description=readme.read_text(),
    long_description_content_type='text/markdown',
    author='KyleTheScientist',
    author_email='kylethescientist@gmail.com',
    url='https://github.com/KyleTheScientist/zOSEdit',
    packages=find_packages(),
    install_requires=[
        'dearpygui',
    ],
    entry_points={'console_scripts': ['zosedit=zosedit.main:main']},
    keywords=['editor', 'z/OS', 'MVS', 'dataset', 'FTP'],
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: Microsoft :: Windows',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



import re
import contextlib
from dearpygui import dearpygui as dpg
from zosedit.gui.dialog import dialog
from zosedit.models import Dataset
from traceback import format_exc
from textwrap import indent


class Explorer:

    def __init__(self, root):
        self.root = root

    def build(self):

        input_options = dict(on_enter=True, callback=self.refresh_datasets, uppercase=True)
        with dpg.group(tag='explorer_search_group'):
            with dpg.tab_bar(tag='explorer_tab_bar', callback=self.on_tab_changed):
                # Datasets tab
                with dpg.tab(label='Datasets', tag='explorer_datasets_tab'):
                    with dpg.group(horizontal=True, tag='explorer_dataset_search_group'):
                        input_width = dpg.get_item_width('win_explorer') - 40
                        dpg.add_input_text(hint='Search',
                                           tag='explorer_dataset_input',
                                           width=input_width,
                                           **input_options)
                        dpg.add_button(label=' O ', callback=self.refresh_datasets)
                    dpg.add_child_window(label='Results', tag='dataset_results')

                # Jobs tab
                input_options['callback'] = self.refresh_jobs
                with dpg.tab(label='Jobs', tag='explorer_jobs_tab'):
                    with dpg.group(tag='explorer_job_search_group', width=-1):
                        dpg.add_input_text(hint='Name', tag='explorer_jobname_input', **input_options)
                        dpg.add_input_text(hint='ID', tag='explorer_jobid_input', **input_options)
                        dpg.add_input_text(hint='Owner', tag='explorer_jobowner_input', **input_options)
                        dpg.add_button(label='Search', callback=self.refresh_jobs)
                    dpg.add_child_window(label='Results', tag='job_results')

        with dpg.theme(tag='explorer_theme_volume'):
            with dpg.theme_component(dpg.mvSelectable):
                dpg.add_theme_color(dpg.mvThemeCol_Text, (170, 170, 170, 255))

        with dpg.theme(tag='rc_theme_error'):
            with dpg.theme_component(dpg.mvSelectable):
                dpg.add_theme_color(dpg.mvThemeCol_Text, (140, 120, 80, 255))

        with dpg.theme(tag='rc_theme_success'):
            with dpg.theme_component(dpg.mvSelectable):
                dpg.add_theme_color(dpg.mvThemeCol_Text, (50, 140, 80, 255))

        with dpg.theme(tag='rc_theme_active'):
            with dpg.theme_component(dpg.mvSelectable):
                dpg.add_theme_color(dpg.mvThemeCol_Text, (70, 100, 160, 255))


    def on_tab_changed(self):
        pass

    def reset(self):
        with self.empty_results('dataset_results'):
            with self.empty_results('job_results'):
                pass
        dpg.set_value('explorer_tab_bar', 'explorer_datasets_tab')
        dpg.set_value('explorer_jobname_input', '')
        dpg.set_value('explorer_jobid_input', '')
        dpg.set_value('explorer_jobowner_input', '')
        dpg.set_value('explorer_dataset_input', '')

    def search_for_job_id(self, id):
        dpg.set_value('explorer_tab_bar', 'explorer_jobs_tab')
        dpg.set_value('explorer_jobname_input', '')
        dpg.set_value('explorer_jobid_input', id)
        dpg.set_value('explorer_jobowner_input', '')
        self.refresh_jobs()

    def refresh_jobs(self):
        name = dpg.get_value('explorer_jobname_input')
        id = dpg.get_value('explorer_jobid_input')
        owner = dpg.get_value('explorer_jobowner_input')
        if not name and not id and not owner:
            return

        # Clear existing results
        with self.empty_results('job_results'):
            # Search for jobs
            status = dpg.add_text('Searching...')
            try:
                jobs = self.root.zftp.list_jobs(name, id, owner)
            except Exception as e:
                dpg.set_value(status, f'Error: {e}')
                print('Error listing jobs')
                print(indent(format_exc(), '    '))
                dpg.configure_item(status, color=(255, 0, 0))
                return
            dpg.set_value(status, f'Found {len(jobs)} job(s)')

            # List results
            with dpg.table(header_row=True, policy=dpg.mvTable_SizingStretchProp):
                dpg.add_table_column(label='ID')
                dpg.add_table_column(label='Name')
                dpg.add_table_column(label='Owner')
                dpg.add_table_column(label='RC')
                for job in jobs:
                    with dpg.table_row():
                        dpg.add_selectable(span_columns=True, label=job.id, callback=self.open_job, user_data=job)
                        dpg.add_selectable(span_columns=True, label=job.name, callback=self.open_job, user_data=job)
                        dpg.add_selectable(span_columns=True, label=job.owner, callback=self.open_job, user_data=job)
                        rc = dpg.add_selectable(span_columns=True, label=job.rc, callback=self.open_job, user_data=job)
                        dpg.bind_item_theme(rc, f'rc_theme_{job.theme()}')

    def refresh_datasets(self):
        # Get datasets
        search = dpg.get_value('explorer_dataset_input')
        if not search:
            return
        if not re.match(r"'[^']+'", search):
            if '*' not in search and len(search.split('.')[-1]) < 8:
                search = f"'{search}*'"
            else:
                search = f"'{search}'"

        with self.empty_results('dataset_results'): # Clears existing results
            # Search for datasets
            status = dpg.add_text('Searching...')
            datasets = [d for d in self.root.zftp.list_datasets(search) if d.type is not None]
            dpg.set_value(status, f'Found {len(datasets)} dataset(s)')

            # List results
            with dpg.table(header_row=True, policy=dpg.mvTable_SizingStretchProp, tag='dataset_results_table'):
                dpg.add_table_column(label='Volume')
                dpg.add_table_column(label='Name')
                for dataset in datasets:
                    self.entry(dataset, leaf=not dataset.is_partitioned())


    def entry(self, dataset: Dataset, leaf: bool, **kwargs):
        with dpg.table_row(parent='dataset_results_table', **kwargs) as row:
            with dpg.table_cell():
                volume = '' if dataset.member else dataset.volume
                selectable = dpg.add_selectable(label=volume, span_columns=True)
                dpg.bind_item_theme(selectable, 'explorer_theme_volume')

            with dpg.table_cell():
                # Create the button/dropdown for the dataset
                name = dataset.name + '/' if dataset.is_partitioned() else dataset.name
                dpg.add_selectable(label=dataset.member or name, span_columns=True)

        # Create context menu
        dpg.popup
        with dpg.window(show=False, autosize=True, popup=True) as context_menu:
            if leaf:
                dpg.add_menu_item(label='Open', callback=self._open_file(dataset))
                dpg.add_menu_item(label='Submit', callback=self._submit_file(dataset))
            else:
                dpg.add_menu_item(label='Create member', callback=self._new_member(dataset))
            dpg.add_menu_item(label='Delete', callback=self.try_delete_file, user_data=dataset)
            dpg.add_menu_item(label='Properties', callback=self.properties_popup, user_data=dataset)

        # Add functionality to the button/dropdown
        with dpg.item_handler_registry() as reg:
            on_left_click = self._open_file(dataset) if leaf else self._populate_pds(dataset, row)
            dpg.add_item_clicked_handler(dpg.mvMouseButton_Left, callback=on_left_click)
            dpg.add_item_clicked_handler(dpg.mvMouseButton_Left, callback=lambda: dpg.set_value(selectable, False))
            dpg.add_item_clicked_handler(dpg.mvMouseButton_Right, callback=lambda: dpg.configure_item(context_menu, show=True))
        dpg.bind_item_handler_registry(selectable, reg)

    def populate_pds(self, dataset: Dataset, parent_row: int):
        if dataset._populated:
            dataset._populated = False
            for child in dpg.get_item_children('dataset_results_table')[1]:
                if dpg.get_item_user_data(child) == dataset:
                    dpg.delete_item(child)
            return

        # Load members
        # status = dpg.add_text('Loading members...', parent=id, indent=10)
        members = self.root.zftp.get_members(dataset)
        # dpg.delete_item(status)

        children = dpg.get_item_children('dataset_results_table')[1]
        index = children.index(parent_row) + 1
        if index < len(children):
            before = children[index]
        else:
            before = 0

        if not members:
            with dpg.table_row(parent='dataset_results_table', before=before, user_data=dataset):
                dpg.add_table_cell()
                with dpg.table_cell():
                    dpg.add_text('<empty>')
            return
        # List members
        for member in members:
            self.entry(dataset=dataset(member), leaf=True, before=before, user_data=dataset)


    def _populate_pds(self, dataset: Dataset, parent: int):
        return lambda: self.populate_pds(dataset, parent)

    def _new_member(self, dataset: Dataset):
        def callback():
            self.root.editor.new_file(dataset.name + '()')
        return callback

    def _open_file(self, dataset: Dataset):
        def callback():
            self.root.editor.open_file(dataset)
        return callback

    def _submit_file(self, dataset: Dataset):
        def callback():
            self.root.zftp.submit_job(dataset)
        return callback

    def open_job(self, sender, data, job):
        dpg.set_value(sender, False)
        self.root.editor.open_job(job)

    def try_delete_file(self, sender, data, dataset):
        w, h = 300, 150
        with dpg.window(modal=True, tag='delete_file_dialog', autosize=True, no_title_bar=True):
            dpg.add_text('Confirm deletion of:', color=(255, 80, 80))
            dpg.add_text(dataset.name, bullet=True)
            with dpg.group(horizontal=True):
                bw = 100
                dpg.add_button(label='Delete', callback=self.delete_file, user_data=dataset, width=bw)
                dpg.add_button(label='Cancel', callback=lambda: dpg.delete_item('delete_file_dialog'), width=bw)

    def delete_file(self, sender, data, dataset):
        dpg.delete_item('delete_file_dialog')
        self.root.zftp.delete(dataset)
        self.root.editor.close_tab_by_dataset(dataset)
        self.refresh_datasets()

    def properties_popup(self, sender, data, dataset):
        with dialog(label=dataset.name, tag='properties_dialog', width=500, height=300):
            properties = dataset.properties()
            with dpg.table(header_row=True, policy=dpg.mvTable_SizingFixedFit):
                dpg.add_table_column(label='Property')
                dpg.add_table_column(label='Value')

                for key, value in properties.items():
                    with dpg.table_row():
                        with dpg.table_cell():
                            dpg.add_text(key.upper())
                        with dpg.table_cell():
                            dpg.add_input_text(readonly=True, default_value=value, width=400)

    @contextlib.contextmanager
    def empty_results(self, item):
        for child in dpg.get_item_children(item)[1]:
            dpg.delete_item(child)

        dpg.push_container_stack(item)
        yield
        dpg.pop_container_stack()
